package no.ntnu.idatt1002.demo.controller;

import java.io.IOException;

import java.util.Optional;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.Node;
import javafx.scene.Parent;
import javafx.scene.Scene;
import javafx.scene.control.Alert;
import javafx.scene.control.Alert.AlertType;
import javafx.scene.control.Button;
import javafx.scene.control.ButtonType;
import javafx.scene.control.ComboBox;
import javafx.scene.control.Dialog;
import javafx.scene.control.TableColumn;
import javafx.scene.control.TableView;
import javafx.scene.control.cell.PropertyValueFactory;
import javafx.scene.text.Text;
import javafx.stage.Modality;
import javafx.stage.Stage;
import no.ntnu.idatt1002.demo.data.Economics.Expense;
import no.ntnu.idatt1002.demo.data.Economics.ExpenseCategory;
import no.ntnu.idatt1002.demo.data.Economics.ExpenseRegister;
import no.ntnu.idatt1002.demo.data.Economics.FileHandling;
import no.ntnu.idatt1002.demo.data.Economics.Income;
import no.ntnu.idatt1002.demo.data.Economics.IncomeRegister;
import no.ntnu.idatt1002.demo.data.Economics.Item;
import no.ntnu.idatt1002.demo.data.Economics.ItemRegister;

/**
 * Controller for expense scene in the application. This controller manages all actions that relates to the expense tableview (add, edit and delete), the switching
 * of scenes from the expense scene to another scene, and the saving of the table, whenever the user switches to another scene.
 *
 * @author Harry Linrui Xu
 * @since 13.3.2023
 */
public class ExpensesController {

  @FXML
  private Button addBtn;
  @FXML
  private Button editBtn;

  @FXML
  private Button deleteBtn;

  @FXML
  private ComboBox<String> show;

  @FXML
  private Button incomeBtn;

  @FXML
  private Text sum;

  @FXML
  private Button budgetBtn;

  @FXML
  private Button returnBtn;


  @FXML
  private TableColumn<Expense, Double> amountColumn;

  @FXML
  private TableColumn<Expense, ExpenseCategory> categoryColumn;

  @FXML
  private TableColumn<Expense, String> dateColumn;

  @FXML
  private TableColumn<Expense, String> descriptionColumn;

  @FXML
  private TableColumn<Expense, Boolean> recurringColumn;

  @FXML
  private TableView<Expense> expenseTableView;

  @FXML
  ExpenseRegister expenseRegister;
  ObservableList<Expense> expenses;

  ObservableList<String> filter;

  /**
   * Initializes the expense register, the observable expense list and the tableview, along values of the dropbox used for filtering the tableview.
   * The method is called each time the FXML of this scene is loaded.
   * @throws IOException If there occurs any exception when loading the budget register from a file.
   */
  @FXML
  public void initialize() throws IOException {
    //Initialize table columns
    dateColumn.setCellValueFactory(new PropertyValueFactory<Expense, String>("date"));
    amountColumn.setCellValueFactory(new PropertyValueFactory<Expense, Double>("amount"));
    categoryColumn.setCellValueFactory(new PropertyValueFactory<Expense, ExpenseCategory>("category"));
    descriptionColumn.setCellValueFactory(new PropertyValueFactory<Expense, String>("description"));
    recurringColumn.setCellValueFactory(new PropertyValueFactory<Expense, Boolean>("recurring"));

    //Initialize the filter box
    filter = FXCollections.observableArrayList("All", "Food", "Clothes", "Books", "Other",
        "Fixed expense");
    show.setItems(filter);
    show.setValue("All");

    //Initialize registers and tableview
    expenseRegister = loadExpenseDataFromFile("Expense");
    expenses = FXCollections.observableArrayList(expenseRegister.getItems());
    expenseTableView.setItems(expenses);

    //Initialize sum field under the tableview
    sum.setText(String.valueOf(expenseRegister.getTotalSum()));
  }

  /**
   * Method for handling the adding of new entries in the tableview.
   * @param event A button click on the add button.
   */
  @FXML
  protected void handleAddButton(ActionEvent event) {
    handleEditButton(event);
  }

  /**
   * Method for handling the editing of a chosen entry in the tableview.
   * @param event A button click on the edit button.
   */
  @FXML
  protected void handleEditButton(ActionEvent event) {
    //Instantiate FXML loader and loads the popup for adding expense
    FXMLLoader loader = new FXMLLoader();
    loader.setLocation(getClass().getResource("/view/AddExpense.fxml"));
    
    Expense newExpense = null;
    String dialogTitle = "";
    // Load the FXML file for your dialog box
    Dialog<Expense> dialog = new Dialog<>();
    dialog.initModality(Modality.APPLICATION_MODAL);

    try {
      // Set the Dialog's content to the loaded FXML file
      dialog.getDialogPane().setContent(loader.load());
    } catch (IOException e) {
      e.printStackTrace();
    }

    // Get the controller for the loaded FXML file
    AddExpenseController dialogController = loader.getController();

    /**
     * The mode of the dialog. NEW if new contact, EDIT if edit existing contact.
     */
    DialogMode dialogMode;    
    //Sets the title of the dialog box
    if (event.getSource().equals(addBtn)) {
      dialogMode = DialogMode.ADD;
      dialogTitle = "Add expense";


    } else if (event.getSource().equals(editBtn)
        && expenseTableView.getSelectionModel().getSelectedItem() != null) {
      dialogMode = DialogMode.EDIT;
      dialogTitle = "Edit expense";
      //Gets the selected item from the table
      newExpense = expenseTableView.getSelectionModel().getSelectedItem();
      //Binds the selected item to another item which is defined in the ItemController
      dialogController.setExpense(newExpense);
    } else {
      return;
    }
    dialog.setTitle(dialogTitle);
    // Show the Dialog and wait for the user to close it
    dialog.showAndWait();
    
    //Get the newly created expense from the dialog pane
    newExpense = dialogController.getNewExpense();

    //Adds the new item to the register
    if (newExpense != null && dialogMode == DialogMode.ADD) {
      expenseRegister.addItem(newExpense);
    }
    //Updates the tableview using the register
    refreshTableView();
  }

  /**
   * Deletes an entry from the tableview, if an entry has been selected. The method brings up a popup window, asking for confirmation for deleting the entry.
   * @param event A button click on the delete button
   */
  @FXML
  public void handleDeleteBtn(ActionEvent event) {
    Expense chosenExpense = expenseTableView.getSelectionModel().getSelectedItem();
    if (chosenExpense == null) {
      return;
    }
    Optional<ButtonType> isConfirmed = showConfirmationDialog();
    if (isConfirmed.isPresent() && isConfirmed.get() == ButtonType.OK) {
      expenseRegister.removeItem(chosenExpense);
      refreshTableView();
    }
  }

  /**
   * Method for synching the register with the tableview. The observable list to which the tableview is set, is being refilled with all the entries
   * in the register, keeping it updated with new changes.
   */
  protected void refreshTableView() {
    this.expenses.setAll(expenseRegister.getItems());
    this.sum.setText(String.valueOf(expenseRegister.getTotalSum()));

  }

  /**
   * Returns an optional, which is a popup alert box, asking for confirmation for deleting an entry.
   * @return An alert box, asking for confirmation for deleting the selected entry of the tableview.
   */
  private Optional<ButtonType> showConfirmationDialog() {
    Alert alert = new Alert(AlertType.CONFIRMATION);
    alert.setTitle("Confirm Delete");
    alert.setHeaderText("Delete Confirmation");
    alert.setContentText("Are you sure you would like to delete the selected entry?");

    return alert.showAndWait();
  }

  /**
   * Method that either reads data from a file with which it fills an expense register, if older changes exist, or instantiates an expense register if the file is empty.
   * @param fileName The name of the file that is being read from.
   * @return An object of type IncomeRegister.
   * @throws IOException If an error occurs while reading from the file.
   */
  public ExpenseRegister loadExpenseDataFromFile(String fileName) throws IOException {
    //ItemRegister<T extends Item>
    FileHandling fileHandling = new FileHandling();
    if (fileHandling.isEmpty(fileName)) {
      expenseRegister = new ExpenseRegister();
    } else {
      try {
        expenseRegister = fileHandling.readExpenseRegisterFromFile(fileName);
      } catch (IOException e) {
        e.printStackTrace();
      }
    }
    return expenseRegister;
  }

  /**
   * Saves the changes made to the expense tableview by writing the information to a file.
   * @param fileName The name of the file that is written to.
   * @throws IOException If an error occurs while writing to the file.
   */
  public void saveDataToFile(String fileName) throws IOException {
    FileHandling fileHandling = new FileHandling();
    fileHandling.writeItemRegisterToFile(expenseRegister, fileName);
  }

  /**
   * Switches scenes from the expense scene to another, by loading a new FXML file and setting the scene to this location.
   * The destination depends entirely on which button is pressed.
   * @param event A button click on the buttons on the buttonbar or the next button
   * @throws IOException If an error occurs with loading any of the FXML files.
   */
  @FXML
  public void switchScene(ActionEvent event) throws IOException {
    saveDataToFile("Expense");
    FXMLLoader loader = new FXMLLoader();
    if (event.getSource() == incomeBtn) {
      loader.setLocation(SceneController.class.getResource("/view/Income.fxml"));
    } else if (event.getSource() == returnBtn) {
      loader.setLocation(SceneController.class.getResource("/view/FirstMenu.fxml"));
    } else if (event.getSource() == budgetBtn) {
      loader.setLocation(SceneController.class.getResource("/view/BudgetNew.fxml"));
    }
      Parent root = loader.load();
      Stage stage = (Stage) ((Node) event.getSource()).getScene().getWindow();
      Scene scene = new Scene(root);
      stage.setScene(scene);
      stage.show();
    }
  }
